#include "mus.h"
#include "gridder.h"

double GetDefaultGapOpenHi();
double GetDefaultGapExtHi();
	
static const vector<double> *g_ptrQs;
// Descending order so > rather than <.
bool LT(uns i, uns j)
	{
	return (*g_ptrQs)[i] > (*g_ptrQs)[j];
	}

void InitSortOrder(vector<uns> &SortOrder, uns N)
	{
	SortOrder.clear();
	SortOrder.reserve(N);
	for (uns i = 0; i < N; ++i)
		SortOrder.push_back(i);
	}

void GetSortOrder(const vector<double> &m_Qs, vector<uns> &SortOrder)
	{
	InitSortOrder(SortOrder, SIZE(m_Qs));
	g_ptrQs = &m_Qs;
	sort(SortOrder.begin(), SortOrder.end(), LT);
	}

//// compute i^j
//static uns ipow(uns i, uns j)
//	{
//	uns prod = 1;
//	for (uns k = 0; k < j; ++k)
//		prod *= i;
//	return prod;
//	}

Gridder::Gridder()
	{
	m_ActiveParamCount = ParamVec::GetActiveParamCount();
	m_fCSV = 0;
	}

uns Gridder::GetCellCount() const
	{
	asserta(m_ActiveParamCount > 0);
	asserta(SIZE(m_GridSizes) == m_ActiveParamCount);
	if (m_ActiveParamCount == 0)
		return 0;
	uns CellCount = 1;
	for (uns i = 0; i < m_ActiveParamCount; ++i)
		CellCount *= m_GridSizes[i];
	return CellCount;
	}

// Coords[i] = 0, ... (N-1)
// i is active parameter index.
void Gridder::CellIndexToCoords(uns CellIndex, vector<uns> &Coords) const
	{
	Coords.clear();
	uns k = CellIndex;
	for (uns i = 0; i < m_ActiveParamCount; ++i)
		{
		uns GridSize = m_GridSizes[i];
		Coords.push_back(k%GridSize);
		k /= GridSize;
		}
	}

double Gridder::GetParamStep(uns ActiveParamIndex) const
	{
	uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
	double lo = m_Lo.Get(ParamIndex);
	double hi = m_Hi.Get(ParamIndex);
	uns N = m_GridSizes[ActiveParamIndex];
	double Step = N == 1 ? 0 : (hi - lo)/(N - 1);
	return fabs(Step);
	}

uns Gridder::CoordsToCellIndex(const vector<uns> &Coords) const
	{
	uns k = 1;
	uns N = SIZE(Coords);
	uns CellIndex = 0;
	for (uns i = 0; i < N; ++i)
		{
		CellIndex += k*Coords[i];
		k *= m_GridSizes[i];
		}
	return CellIndex;
	}

void Gridder::CellIndexToParams(uns CellIndex, ParamVec &Params) const
	{
	Params.Zero(1.0);

	for (uns ActiveParamIndex = 0; ActiveParamIndex < m_ActiveParamCount;
	  ++ActiveParamIndex)
		{
		uns GridSize = m_GridSizes[ActiveParamIndex];
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		SCORE LoParam = m_Lo.Get(ParamIndex);
		SCORE HiParam = m_Hi.Get(ParamIndex);
		uns Cell = CellIndex%GridSize;
		double w = GridSize == 1 ? 0 : (double) Cell/(double) (GridSize - 1);
		SCORE Param = SCORE(LoParam + (HiParam - LoParam)*w);
		Params.Set(ParamIndex, Param);
		CellIndex /= GridSize;
		}
	}

double Gridder::GetPlusGradient(uns CellIndex, uns ActiveParamIndex) const
	{
	vector<uns> Coords;
	CellIndexToCoords(CellIndex, Coords);

	double Q = m_Qs[CellIndex];
	uns StartCoord = Coords[ActiveParamIndex];
	const uns GridSize = m_GridSizes[ActiveParamIndex];
	for (uns Coord = StartCoord + 1; Coord < GridSize; ++Coord)
		{
		Coords[ActiveParamIndex] = (uns) Coord;
		uns CellIndex2 = CoordsToCellIndex(Coords);
		double Q2 = m_Qs[CellIndex2];
		if (Q2 != Q)
			{
			double dQ = Q2 - Q;
			double dParam = GetParamStep(ActiveParamIndex)*(Coord - StartCoord);
			return dQ/dParam;
			}
		}
	return 0.0;
	}

uns Gridder::GetPlusDeltaCell(uns CellIndex, uns ActiveParamIndex) const
	{
	vector<uns> Coords;
	CellIndexToCoords(CellIndex, Coords);

	double Q = m_Qs[CellIndex];
	uns StartCoord = Coords[ActiveParamIndex];
	const uns GridSize = m_GridSizes[ActiveParamIndex];
	for (uns Coord = StartCoord + 1; Coord < GridSize; ++Coord)
		{
		Coords[ActiveParamIndex] = (uns) Coord;
		uns CellIndex2 = CoordsToCellIndex(Coords);
		double Q2 = m_Qs[CellIndex2];
		if (Q2 != Q)
			return CellIndex2;
		}
	return UINT_MAX;
	}

double Gridder::GetAvgGradient(uns CellIndex, uns ActiveParamIndex) const
	{
	double gplus = GetPlusGradient(CellIndex, ActiveParamIndex);
	double gminus = GetMinusGradient(CellIndex, ActiveParamIndex);
	return (fabs(gplus) + fabs(gminus))/2;
	}

void Gridder::GetGradient(uns CellIndex, ParamVec &Grads) const
	{
	Grads.Zero(1.0);
	for (uns ActiveParamIndex = 0; ActiveParamIndex < m_ActiveParamCount;
	  ++ActiveParamIndex)
		{
		double g = GetAvgGradient(CellIndex, ActiveParamIndex);
		uns ParamIndex = ParamVec::ActiveIndexToIndex(ActiveParamIndex);
		Grads.Set(ParamIndex, (SCORE) g);
		}
	}

double Gridder::GetMinusGradient(uns CellIndex, uns ActiveParamIndex) const
	{
	vector<uns> Coords;
	CellIndexToCoords(CellIndex, Coords);

	double Q = m_Qs[CellIndex];
	uns StartCoord = Coords[ActiveParamIndex];
	for (int Coord = (int) StartCoord - 1; Coord >= 0; --Coord)
		{
		Coords[ActiveParamIndex] = (uns) Coord;
		uns CellIndex2 = CoordsToCellIndex(Coords);
		double Q2 = m_Qs[CellIndex2];
		if (Q2 != Q)
			{
			double dQ = Q - Q2;
			double dParam = GetParamStep(ActiveParamIndex)*(StartCoord - Coord);
			return dQ/dParam;
			}
		}
	return 0.0;
	}

uns Gridder::GetMinusDeltaCell(uns CellIndex, uns ActiveParamIndex) const
	{
	vector<uns> Coords;
	CellIndexToCoords(CellIndex, Coords);

	double Q = m_Qs[CellIndex];
	uns StartCoord = Coords[ActiveParamIndex];
	for (int Coord = (int) StartCoord - 1; Coord >= 0; --Coord)
		{
		Coords[ActiveParamIndex] = (uns) Coord;
		uns CellIndex2 = CoordsToCellIndex(Coords);
		double Q2 = m_Qs[CellIndex2];
		if (Q2 != Q)
			return CellIndex2;
		}
	return UINT_MAX;
	}

bool Gridder::IsPartialLocalOptimum(uns CellIndex, uns ActiveParamIndex) const
	{
	return GetMinusGradient(CellIndex, ActiveParamIndex) > 0 &&
	  GetPlusGradient(CellIndex, ActiveParamIndex) < 0;
	}

bool Gridder::IsLocalOptimum(uns CellIndex) const
	{
	for (uns ActiveParamIndex = 0; ActiveParamIndex < m_ActiveParamCount;
	  ++ActiveParamIndex)
		if (!IsPartialLocalOptimum(CellIndex, ActiveParamIndex))
			return false;

	return true;
	}

void Gridder::GetLocalOptima(vector<uns> &CellIndexes) const
	{
	CellIndexes.clear();
	const uns CellCount = GetCellCount();
	asserta(SIZE(m_Qs) == CellCount);
	
	for (uns CellIndex = 0; CellIndex < CellCount; ++CellIndex)
		if (IsLocalOptimum(CellIndex))
			CellIndexes.push_back(CellIndex);
	}

void Gridder::ComputeGrid(ptrGetQFn GetQ, const char *CSVFileName)
	{
	if (CSVFileName != 0)
		m_fCSV = CreateStdioFile(CSVFileName);

	m_Qs.clear();
	const uns CellCount = GetCellCount();

	Log("\n");
	Log("Grid, %u cells\n", CellCount);
	
	Log("   Cell");
	Log("        Q");
	ParamVec::LogRowHdr1();
	if (ParamVec::IsActive(P_dx))
		Log("    DAL");
	Log("\n");

	Log("-------");
	Log("  -------");
	ParamVec::LogRowHdr2();
	if (ParamVec::IsActive(P_dx))
		Log("  -----");
	Log("\n");
	double BestQ = 0;
	ParamVec BestParams;
	for (uns CellIndex = 0; CellIndex < CellCount; ++CellIndex)
		{
		ParamVec Params;
		Progress("Cell %u of %u %.1f%%", CellIndex+1, CellCount, Pct(CellIndex+1, CellCount));
		CellIndexToParams(CellIndex, Params);
		m_ParamVecs.push_back(Params);
		double Q = GetQ(Params);
		m_Qs.push_back(Q);

		if (Q > BestQ)
			{
			BestQ = Q;
			BestParams = Params;
			}

		Log("%7u  %7.5f", CellIndex, Q);
		Params.LogRow();
		if (ParamVec::IsActive(P_dx))
			Log("  %5u", Params.GetDAL());
		Log("\n");

		if (m_fCSV != 0)
			{
			string CSV;
			Params.ToSepStr(CSV, ';');
			fprintf(m_fCSV, "%s;Q=%.4f\n", CSV.c_str(), Q);
			}
		}

	if (m_fCSV != 0)
		{
		CloseStdioFile(m_fCSV);
		m_fCSV = 0;
		}

	Log("\n");
	Log("BestQ=%.4f BestParams=", BestQ);
	BestParams.LogMe(true);
	}

void Gridder::LogCells(const vector<uns> &CellIndexes) const
	{
	Log("\n");
	Log("   Cell");
	Log("        Q");
	ParamVec::LogRowHdr1();
	ParamVec::LogRowHdr1('+');
	ParamVec::LogRowHdr1('+');
	ParamVec::LogRowHdr1('-');
	ParamVec::LogRowHdr1('-');
	Log("\n");

	Log("-------");
	Log("  -------");
	ParamVec::LogRowHdr2();
	ParamVec::LogRowHdr2();
	ParamVec::LogRowHdr2();
	ParamVec::LogRowHdr2();
	ParamVec::LogRowHdr2();
	Log("\n");

	const uns N = SIZE(CellIndexes);

	vector<uns> SortedCellIndexes(CellIndexes);
	SortCellIndexesByQ(SortedCellIndexes);

	for (uns i = 0; i < N; ++i)
		{
		uns CellIndex = SortedCellIndexes[i];
		Log("%7u  %7.5f", CellIndex, m_Qs[CellIndex]);
		const ParamVec &Params = m_ParamVecs[CellIndex];
		Params.LogRow();

		for (uns API = 0; API < m_ActiveParamCount; ++API)
			{
			uns CellIndex2 = GetPlusDeltaCell(CellIndex, API);
			if (CellIndex2 == UINT_MAX)
				Log("          *", CellIndex2);
			else
				Log("  %9u", CellIndex2);
			}

		for (uns API = 0; API < m_ActiveParamCount; ++API)
			{
			double grad = GetPlusGradient(CellIndex, API);
			Log("  %9.5f", grad);
			}

		for (uns API = 0; API < m_ActiveParamCount; ++API)
			{
			uns CellIndex2 = GetMinusDeltaCell(CellIndex, API);
			if (CellIndex2 == UINT_MAX)
				Log("          *", CellIndex2);
			else
				Log("  %9u", CellIndex2);
			}

		for (uns API = 0; API < m_ActiveParamCount; ++API)
			{
			double grad = GetMinusGradient(CellIndex, API);
			Log("  %9.5f", grad);
			}

		Log("\n");
		}
	}

void Gridder::SortCellIndexesByQ(vector<uns> &CellIndexes) const
	{
	g_ptrQs = &m_Qs;
	sort(CellIndexes.begin(), CellIndexes.end(), LT);
	}

void Gridder::InitFromCmdLine()
	{
	uns GridSize = 8;
	IntOpt("gridsize", &GridSize);
	const uns ActiveParamCount = ParamVec::GetActiveParamCount();
	m_GridSizes.resize(ActiveParamCount, GridSize);

	m_Lo.Zero(1.0);
	m_Hi.Zero(1.0);

	for (uns i = 0; i < ActiveParamCount; ++i)
		{
		uns ParamIndex = ParamVec::ActiveIndexToIndex(i);
		string ParamName = AlnParamToStr(ParamIndex);
		string Namelo = ParamName + "lo";
		string Namehi = ParamName + "hi";
		string nName = "n" + ParamName;
		double lo = 0;
		double hi = 0;
		if (ParamName == "g" || ParamName == "tg")
			{
			hi = GetDefaultGapOpenHi();
			if (ValueOpt(Namehi.c_str()) == 0)
				Log("Gap open hi %g\n", hi);
			}
		else if (ParamName == "e" || ParamName == "te")
			{
			hi = GetDefaultGapExtHi();
			if (ValueOpt(Namehi.c_str()) == 0)
				Log("Gap ext hi %g\n", hi);
			}

		FloatOpt(Namehi.c_str(), &hi);
		FloatOpt(Namelo.c_str(), &lo);

		if (ParamVec::IsNegativeParam(ParamIndex))
			{
			lo = -lo;
			hi = -hi;
			}

		m_Lo.Set(ParamIndex, (SCORE) lo);
		m_Hi.Set(ParamIndex, (SCORE) hi);

		uns n = GridSize;
		IntOpt(nName.c_str(), &n);
		m_GridSizes[i] = n;
		}

	Log("\n");
	Log("%u active parameters\n", ActiveParamCount);
	Log("Param         Lo         Hi       N       Step\n");
	Log("-----  ---------  ---------  ------  ---------\n");
	for (uns i = 0; i < ActiveParamCount; ++i)
		{
		uns j = ParamVec::ActiveIndexToIndex(i);
		Log("%5.5s", AlnParamToStr(j));
		double lo = m_Lo.Get(j);
		double hi = m_Hi.Get(j);
		Log("  %9.5g", lo);
		Log("  %9.5g", hi);
		uns N = m_GridSizes[i];
		Log("  %6u", N);
		double Step = GetParamStep(i);
		Log("  %9.5g\n", Step);
		}
	}
