#include "mus.h"
#include "readfasta.h"

double GetWeightFromLabel(const string &Label);

static MSASet *g_ptrMSASet;
static string g_GroupLabel;
static vector<string> g_MSA;
static vector<string> g_Labels;

static void GetUngappedSeq(const string &s, string &t)
	{
	t.clear();
	for (string::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		char c = *p;
		if (!IsGapChar(c))
			t.push_back(c);
		}
	}

void GetUngappedSeqs(const vector<string> &MSA, vector<string> &Seqs)
	{
	Seqs.clear();
	const uns N = SIZE(MSA);
	for (uns i = 0; i < N; ++i)
		{
		string Empty;
		Seqs.push_back(Empty);
		string &Seq = Seqs.back();
		GetUngappedSeq(MSA[i], Seq);
		}
	}

static void OnSeq(const string &Label, const string &Seq)
	{
	if (Seq.empty())
		return;
	g_Labels.push_back(Label);
	if (!g_MSA.empty() && SIZE(Seq) != SIZE(g_MSA.front()))
		Die("Group %s, seqs not aligned", g_GroupLabel.c_str());
	g_MSA.push_back(Seq);
	}

static void OnGroupLabel(const string &Label)
	{
	if (!g_MSA.empty())
		{
		g_ptrMSASet->AddMSA(g_GroupLabel, g_MSA, g_Labels);
		g_MSA.clear();
		g_Labels.clear();
		}
	g_GroupLabel = Label;
	}

void MSASet::Clear()
	{
	m_RefMSAs.clear();
	m_UngappedSeqs.clear();
	m_LabelVec.clear();
	m_RefLabelVec.clear();
	m_Names.clear();
	}

void MSASet::FromFile(const char *FileName)
	{
	Clear();
	g_ptrMSASet = this;
	ReadFasta(FileName, OnSeq, OnGroupLabel);
	OnGroupLabel("");
	}

void MSASet::AddMSA(const string &Name, const vector<string> &MSA,
  const vector<string> &Labels)
	{
	m_Names.push_back(Name);
	m_LabelVec.push_back(Labels);

	vector<string> Seqs;
	::GetUngappedSeqs(MSA, Seqs);
	m_UngappedSeqs.push_back(Seqs);

	m_RefMSAs.push_back(MSA);
	m_RefLabelVec.push_back(Labels);
	}

void MSASet::GetAllPairs(PWSet &AllPairs, bool MemSave) const
	{
	AllPairs.Clear();
	AllPairs.SetMemSave(MemSave);
	uns N = GetCount();
	ProgressStart("Get all pairs");
	for (uns MSAIndex = 0; MSAIndex < N; ++MSAIndex)
		{
		if (MSAIndex%10 == 0)
			ProgressStep(MSAIndex, N);
		const vector<string> &MSA = m_RefMSAs[MSAIndex];
		const vector<string> &Labels = m_LabelVec[MSAIndex];
		const string &MSAName = m_Names[MSAIndex];
		const uns SeqCount = SIZE(MSA);
		asserta(SeqCount >= 2);
		const uns PairCount = (SeqCount*(SeqCount-1))/2;
		double Weight = 1.0/PairCount;
		for (uns SeqIndex1 = 0; SeqIndex1 < SeqCount; ++SeqIndex1)
			{
			const string &Seq1 = MSA[SeqIndex1];
			if (MemSave && Seq1.find('+') != string::npos)
				Die("-memsave with truncated seqs");
			const string &Label1 = Labels[SeqIndex1];

			char Tmp[16];
			sprintf(Tmp, "%u", SeqIndex1);
			string S1(Tmp);

			for (uns SeqIndex2 = SeqIndex1+1; SeqIndex2 < SeqCount; ++SeqIndex2)
				{
				sprintf(Tmp, "%u", SeqIndex2);
				string S2(Tmp);

				const string &Seq2 = MSA[SeqIndex2];
				const string &Label2 = Labels[SeqIndex2];
				string Name = MSAName;
				if (SeqCount == 2)
					Weight = GetWeightFromLabel(Name);
				else
					Name += "." + S1 + "." + S2;

				AllPairs.AddPair(Name, Label1, Seq1, Label2, Seq2, Weight,
				  MSAIndex, SeqIndex1, SeqIndex2);
				}
			}
		}
	ProgressDone();
	Progress("%u pairs", AllPairs.GetPairCount());
	AllPairs.Init();
	}

void MSASet::GetPWSet(const vector<uns> &MSAIndexes,
  const vector<pair<uns,uns> > &SeqIndexes, const vector<double> &PairWeights,
  PWSet &Set) const
	{
	Set.Clear();
	const uns N = SIZE(MSAIndexes);
	asserta(SIZE(PairWeights) == N);
	asserta(SIZE(SeqIndexes) == N);
	for (uns i = 0; i < N; ++i)
		{
		uns MSAIndex = MSAIndexes[i];
		asserta(MSAIndex < SIZE(m_RefMSAs));
		const vector<string> &MSA = m_RefMSAs[MSAIndex];

		uns SeqIndex1 = SeqIndexes[i].first;
		uns SeqIndex2 = SeqIndexes[i].second;
		const uns SeqCount = SIZE(MSA);
		asserta(SeqIndex1 < SeqCount);
		asserta(SeqIndex2 < SeqCount);

		const string &Label1 = m_LabelVec[MSAIndex][SeqIndex1];
		const string &Label2 = m_LabelVec[MSAIndex][SeqIndex2];

		const string &Seq1 = MSA[SeqIndex1];
		const string &Seq2 = MSA[SeqIndex2];

		double PairWeight = PairWeights[i];

		char Tmp[16];
		sprintf(Tmp, "%u", SeqIndex1);
		string S1(Tmp);
		sprintf(Tmp, "%u", SeqIndex2);
		string S2(Tmp);
		string Name = m_Names[MSAIndex] + "." + S1 + "." + S2;
		Set.AddPair(Name, Label1, Seq1, Label2, Seq2, PairWeight,
		  SeqIndex1, SeqIndex2);
		}
	Set.Init();
	}

uns MSASet::GetSeqCount() const
	{
	uns Sum = 0;
	uns Count = GetCount();
	for (uns i = 0; i < Count; ++i)
		Sum += GetSeqCount(i);
	return Sum;
	}

uns MSASet::GetPairCount() const
	{
	uns Sum = 0;
	uns Count = GetCount();
	for (uns i = 0; i < Count; ++i)
		Sum += GetPairCount(i);
	return Sum;
	}

uns MSASet::GetPairCount(uns MSAIndex) const
	{
	uns N = GetSeqCount(MSAIndex);
	return (N*(N-1))/2;
	}

uns MSASet::GetSeqCount(uns MSAIndex) const
	{
	asserta(MSAIndex < SIZE(m_UngappedSeqs));
	return SIZE(m_UngappedSeqs[MSAIndex]);
	}

static void GetRandomIndexes(uns N, vector<uns> &Indexes)
	{
	Indexes.clear();
	for (uns i = 0; i < N; ++i)
		Indexes.push_back(i);
	RandomizeVec(Indexes);
	}

void MSASet::GetSubset(MSASet &Set, uns MaxMSAs, uns MaxSeqsPerMSA) const
	{
	Set.Clear();
	uns N = GetCount();
	if (MaxMSAs < N)
		N = MaxMSAs;

	vector<uns> MSAIndexes;
	for (uns i = 0; i < N; ++i)
		MSAIndexes.push_back(i);
	RandomizeVec(MSAIndexes);

	for (uns i = 0; i < N; ++i)
		{
		uns MSAIndex = MSAIndexes[i];
		const string &Name = GetName(MSAIndex);

		const vector<string> &Labels = GetLabels(MSAIndex);
		const vector<string> &Seqs = GetRefMSA(MSAIndex);
		const uns SeqCount = SIZE(Seqs);
		asserta(SIZE(Labels) == SeqCount);
		if (SeqCount <= MaxSeqsPerMSA)
			{
			Set.AddMSA(Name, Seqs, Labels);
			continue;
			}

		const uns M = min(MaxSeqsPerMSA, SeqCount);
		vector<uns> Indexes;
		GetRandomIndexes(M, Indexes);
		asserta(SIZE(Indexes) == M);
		asserta(M <= SeqCount);

		vector<string> Seqs2;
		vector<string> Labels2;
		for (uns j = 0; j < M; ++j)
			{
			uns Index = Indexes[j];
			Seqs2.push_back(Seqs[Index]);
			Labels2.push_back(Labels[Index]);
			}
		Set.AddMSA(Name, Seqs2, Labels2);
		}
	}

void MSASet::ToFile(const char *FileName) const
	{
	FILE *f = CreateStdioFile(FileName);
	const uns Count = GetCount();
	for (uns MSAIndex = 0; MSAIndex < Count; ++MSAIndex)
		MSAToFile(f, MSAIndex);
	CloseStdioFile(f);
	}

uns MSASet::GetMaxSeqLength(uns MSAIndex) const
	{
	const vector<string> &Seqs = m_UngappedSeqs[MSAIndex];
	const uns SeqCount = SIZE(Seqs);
	uns MaxLength = 0;
	for (uns i = 0; i < SeqCount; ++i)
		{
		const string &Seq = Seqs[i];
		if (SIZE(Seq) > MaxLength)
			MaxLength = SIZE(Seq);
		}
	return MaxLength;
	}

void MSASet::MSAToFile(FILE *f, uns MSAIndex) const
	{
	fprintf(f, "<%s\n", GetName(MSAIndex).c_str());

	asserta(MSAIndex < SIZE(m_RefMSAs));
	asserta(MSAIndex < SIZE(m_LabelVec));
	const vector<string> &Seqs = m_RefMSAs[MSAIndex];
	const vector<string> &Labels = m_LabelVec[MSAIndex];

	const uns SeqCount = SIZE(Seqs);
	asserta(SIZE(Labels) == SeqCount);
	for (uns i = 0; i < SeqCount; ++i)
		{
		fprintf(f, ">%s\n", Labels[i].c_str());

		const string &Seq = Seqs[i];
		WriteFasta(f, Seq.c_str(), SIZE(Seq));
		}
	}

void MSASet::GetPWSubset(uns SubsetSize, PWSet &P) const
	{
	uns MSACount = GetCount();
	uns PairCount = GetPairCount();

	if (SubsetSize == UINT_MAX || SubsetSize >= PairCount)
		{
		GetAllPairs(P);
		return;
		}

	uns MaxPairsPerMSA = 1;
	for (;;)
		{
		uns SubsetPairCount = 0;
		for (uns i = 0; i < MSACount; ++i)
			{
			const uns SeqCount = GetSeqCount(i);
			uns PairCount_i = (SeqCount*(SeqCount - 1))/2;
			if (PairCount_i > MaxPairsPerMSA)
				PairCount_i = MaxPairsPerMSA;
			SubsetPairCount += PairCount_i;
			}
		if (SubsetPairCount > SubsetSize)
			{
			if (MaxPairsPerMSA > 1)
				--MaxPairsPerMSA;
			break;
			}
		++MaxPairsPerMSA;
		}

	vector<uns> Indexes;
	for (uns i = 0; i < MSACount; ++i)
		Indexes.push_back(i);
	RandomizeVec(Indexes);

	vector<uns> MSAIndexes;
	vector<double> PairWeights;
	vector<pair<uns, uns> > SeqIndexes;
	uns TotalPairCount = 0;
	for (uns i = 0; i < MSACount; ++i)
		{
		uns MSAIndex = Indexes[i];
		uns SeqCount = GetSeqCount(MSAIndex);
		if (SeqCount == 1)
			Die("Alignment has 1 seq");
		uns PairCount = (SeqCount*(SeqCount - 1))/2;
		if (TotalPairCount + PairCount > SubsetSize)
			PairCount = SubsetSize - TotalPairCount;
		if (PairCount > MaxPairsPerMSA)
			PairCount = MaxPairsPerMSA;
		double Weight = 1./MaxPairsPerMSA;
		vector<pair<uns, uns> > Pairs;
		for (uns i = 0; i < SeqCount; ++i)
			for (uns j = 0; j < i; ++j)
				Pairs.push_back(pair<uns, uns>(i, j));

		vector<uns> PairIndexes;
		for (uns i = 0; i < PairCount; ++i)
			PairIndexes.push_back(i);
		RandomizeVec(PairIndexes);
		for (uns i = 0; i < PairCount; ++i)
			{
			uns k = PairIndexes[i];
			MSAIndexes.push_back(MSAIndex);
			SeqIndexes.push_back(Pairs[k]);
			PairWeights.push_back(Weight);
			}
		TotalPairCount += MaxPairsPerMSA;
		if (TotalPairCount >= SubsetSize)
			break;
		}
	GetPWSet(MSAIndexes, SeqIndexes, PairWeights, P);
	}
