#ifndef mus_h
#define mus_h

#include <cmath>
#include <vector>
#include <set>

using namespace std;

#ifdef _MSC_VER
#pragma warning(disable: 4244)	// double->float
#pragma warning(disable: 4715)	// not all control paths return value
#endif

#include "mustypes.h"
#include "utils.h"
#include "best.h"
#include "alnparam.h"
#include "readfasta.h"
#include "path.h"
#include "getopts.h"
#include "progress.h"
#include "paramvec.h"
#include "pwset.h"
#include "msaset.h"
#include "paramspace.h"
#include "letters.h"
#include "pwa.h"
#include "hillclimber.h"

enum EDGE
	{
	Edge_SS,	// 0
	Edge_SM,	// 1
	Edge_SD,	// 2
	Edge_SX,	// 3
	Edge_SI,	// 4
	Edge_SY,	// 5

	Edge_MS,	// 6
	Edge_MM,	// 7
	Edge_MD,	// 8
	Edge_MI,	// 9
	Edge_MX,	// 10
	Edge_MY,	// 11

	Edge_DS,	// 12
	Edge_DM,	// 13
	Edge_DD,	// 14
	Edge_DI,	// 15

	Edge_XM,	// 16
	Edge_XX,	// 17

	Edge_YM,	// 18
	Edge_YY,	// 19

	Edge_IS,	// 20
	Edge_IM,	// 21
	Edge_ID,	// 22
	Edge_II,	// 23
	};

const char *EdgeToStr(uns Edge);

typedef void (*ptrOnGapFn)(uns ColIndex, uns Length);
typedef void (*ptrOnPathFn)(const string &Path);

const SCORE MINUS_INFINITY = (SCORE) -1e37;
const byte INVALID_LETTER = 0xff;
const SCORE SCORE_INSANE = SCORE(FLT_MAX);

// Are two floats equal to within epsilon?
const double epsilon = 0.01;
inline bool feq(double x, double y, double epsilon)
	{
	if (fabs(x) > 10000)
		epsilon = fabs(x)/10000;
	if (fabs(x - y) > epsilon)
		return false;
	return true;
	}

inline bool feq(double x, double y)
	{
	double e = epsilon;
	if (fabs(x) > 10000)
		e = fabs(x)/10000;
	if (fabs(x - y) > e)
		return false;
	return true;
	}

void ParseStrVec(const std::string &In, std::vector<std::string> &Out, char Sep);
void ParseStrVecWS(const std::string &In, std::vector<std::string> &Out);
void ReadSubstMx(const char *FileName, uns iMx = 1);
void ReadSubstMx(const char *FileName, ScoreMx &Mx);
void SetScoreMx(ScoreMx &Mx, uns LA, uns LB, SCORE v);
void LogSubstMx(uns iMx = 1);
void LogScoreMx(const ScoreMx &Mx, uns LA, uns LB);
ScoreMx GetJointScoreMx();
const SCORE *GetBackScoreVec();
SCORE NW(const vector<byte> &A, const vector<byte> &B, const ParamVec &Params,
  bool DoLoss, string &Path);
SCORE NW2(const vector<byte> &A, const vector<byte> &B, const ParamVec &Params,
  bool DoLoss, string &Path);
SCORE NW6(const vector<byte> &A, const vector<byte> &B, const ParamVec &Params,
  bool ALeftTrunc, bool ARightTrunc, bool BLeftTrunc, bool BRightTrunc, string &Path);
void LogPWAln(const string &A, const string &B, const string &Path,
  bool SubAln = false);
void AlnStr(const string &A, const string &Path, bool First, string &sA,
  bool ValidateLength = true);
void StripGaps(const vector<byte> &s, vector<byte> &t);
void StripGaps(const string &s, string &t);
void QScore(const char *RefName, 
  const vector<string> &msaTest, const vector<string> &TestLabels,
  const vector<string> &msaRef,  const vector<string> &RefLabels,
  double &Q, double &TC, bool WarnIfNoRefAligned);
unsigned GetAlphabetSize();
void AllocScoreMx(unsigned n, unsigned m, ScoreMx &Mx);
void FreeScoreMx(unsigned n, unsigned m, ScoreMx &Mx);
void SeqsToByteVecs(const vector<string> &GappedSeqs, vector<string> &UngappedSeqs,
  vector<vector<byte> > &GappedBVs, vector<vector<byte> > &UngappedBVs);
void ReadMFA(const char *FileName, vector<string> &Labels, vector<string> &Seqs);
void SetLoss(const string &A, const string &B, double Scale);
SCORE GetFeatureScore(const string &NameA, const string &NameB, const string &A,
  const string &B, const ParamVec &W, bool Dump);
const ScoreMx &GetSubstMx(uns iMx = 1);
void GetSubstMxFromCmdLine(uns iMx = 1);
void GetFeatureVec(const string &A, const string &B, ParamVec &F);
void GetDiverseBest(const vector<ParamVec> &Ws, const vector <double> &Qs, uns k,
  double QPct, vector<uns> &Indexes);
double GetAvgAbsSubstMxAbsValue(uns iMx = 1);
double ComputeQ(const PWSet &Set, const ParamVec &Params);
void Round10(double x, unsigned SigFig, int &Mantissa, int &Exponent);
const char *gstr(double x, unsigned s);
void RandomizeVec(vector<uns> &Vec);
void GetDiverseBest(const vector<ParamVec> &Ws, const vector <double> &Qs, uns k,
  double QPct, vector<uns> &Indexes);
void LogWQs(const vector<ParamVec> &Ws, const vector<double> &Qs);
uint32 RandInt32();
void WriteFasta(FILE *f, const char *SeqData, unsigned Length);
void WriteFasta(FILE *f, const string &Seq);
void WriteFasta(FILE *f, const std::string &Label, const std::string &Seq);
void InitRand(uint32 Seed);
void SeqWeightsFromMSA(const vector<string> &MSA, vector<double> &SeqWeights);
void ScanSeqGaps(const string &Seq, ptrOnGapFn OnGap, ptrOnGapFn OnTerminalGap);
void LogDPMx(const ScoreMx &Mx, uns LA, uns LB);
void LogDPMx(const ScoreMx &Mx, const vector<byte> &A, const vector<byte> &B);
void LogProbMx(const ScoreMx &Mx, const vector<byte> &A, const vector<byte> &B);
void LogTBMx(const CharMx &Mx, uns LA, uns LB);
void LogTBMx(const CharMx &Mx, const vector<byte> &A, const vector<byte> &B);
void SetSubstMxGapScores(ScoreMx &Mx, const ParamVec &Params);
void MSAFromFile(const char *FileName, vector<string> &Seqs,
  vector<string> &Labels);
void WriteMSA(FILE *f, const vector<string> &MSA, const vector<string> &Labels);
void StripGapColsMSA(const vector<string> &MSA, vector<string> &OutSeqs);
void LogMSA(const vector<string> &MSA, const vector<string> &Labels);
void MakeMSA(const string &Path, const vector<string> &MSA1,
  const vector<string> &MSA2, const vector<string> &Labels1,
  const vector<string> &Labels2, vector<string> &MSA, vector<string> &Labels);
bool InitLetters();
void InitSortOrder(vector<uns> &SortOrder, uns N);
void GetSortOrder(const vector<double> &Qs, vector<uns> &SortOrder);
double MSASetGetQ(const MSASet &Set, const ParamVec &Params, uns RefineIters,
  double &TC);
void GetUngappedSeqs(const vector<string> &MSA, vector<string> &Seqs);
double GetHydroChar(char c);
double GetHydroLetter(uns Letter);
double GetHydroMSACol(const vector<string> &MSA, uns ColIndex);
void WindowSmooth(const vector<double> &Hydros, uns WindowLength, bool Dropoff,
  vector<double> &Hydrosw);
bool IsFirstState(char c);
bool IsSecondState(char c);
SCORE SumLog5(SCORE v, SCORE w, SCORE x, SCORE y, SCORE z);
SCORE SumLog3(SCORE x, SCORE y, SCORE z);
SCORE SumLog2(SCORE x, SCORE y);
SCORE safeexp(SCORE x);
uns GetFirstCount(const string &Path);
uns GetSecondCount(const string &Path);

extern uns g_HydroWindowLength;
extern bool g_HydroDropoff;
extern bool g_MAFFT;
extern bool g_LogDP;
extern bool g_LogAlns;
extern bool g_MEA;
extern bool g_Infer;
extern const string *g_LeftInferResults;
extern const string *g_RightInferResults;
extern float **g_Loss;

#endif // mus_h
