#include "mus.h"
#include "gridder.h"

void ComputeGrid(ptrGetQFn GetQ, const ParamVec &Lo, const ParamVec &Hi,
  const vector<uns> &GridSizes, vector<ParamVec> &ParamVecs, vector<double> &Qs);
void GetLocalOptima(const vector<double> &Qs, const vector<uns> &GridSizes,
  vector<uns> &CellIndexes);

static PWSet *g_ptrPWSet;

static double PWGetQ(const ParamVec &Params)
	{
	return g_ptrPWSet->ComputeQ(Params);
	}

void GetSubstMxMinMax(double &Min, double &Max)
	{
	const ScoreMx &Mx = GetSubstMx();
	Min = FLT_MAX;
	Max = -FLT_MAX;
	unsigned AS = GetAlphabetSize();
	for (unsigned i = 0; i < AS; ++i)
		{
		for (unsigned j = 0; j < AS; ++j)
			{
			SCORE s = Mx[i][j];
			if (s < Min)
				Min = s;
			if (s > Max)
				Max = s;
			}
		}
	}

double GetDefaultGapOpenHi()
	{
	double Min;
	double Max;
	GetSubstMxMinMax(Min, Max);
	return 2*(Max - Min);
	}

double GetDefaultGapExtHi()
	{
	return GetDefaultGapOpenHi()/3;
	}

static void FillStartCells(vector<uns> &StartCellIndexes, uns PathCount,
  const vector<uns> &CellIndexesByQ)
	{
	uns CellCount = SIZE(CellIndexesByQ);
	vector<bool> Done(CellCount, false);

	for (uns i = 0; i < SIZE(StartCellIndexes); ++i)
		{
		uns CellIndex = StartCellIndexes[i];
		for (uns j = 0; j < CellCount; ++j)
			{
			if (CellIndexesByQ[j] == CellIndex)
				{
				Done[j] = true;
				break;
				}
			}
		}

	for (;;)
		{
		bool AnyNotDone = false;
		if (SIZE(StartCellIndexes) >= PathCount)
			return;
		for (uns i = 0; i < CellCount; ++i)
			{
			if (Done[i])
				continue;
			uns CellIndex = CellIndexesByQ[i];
			AnyNotDone = true;
			StartCellIndexes.push_back(CellIndex);
			Done[i] = true;
			}
		if (!AnyNotDone)
			return;
		}
	}

void Pop()
	{
	const char *InFileName = RequiredValueOpt("pop");
	uns SubsetSize1 = 20;
	uns SubsetSize2 = 200;
	uns Seed = 1;
	uns PathCount = 4;
	uns MaxDirs2 = 2;
	uns MaxDirs3 = 2;

    double Mindq2 = 0.001;
    double Maxdq2 = 0.01;
    double Mindp2 = 0.01;
    double Maxdp2 = 1.0;

    double Mindq3 = 0.001;
    double Maxdq3 = 0.003;
    double Mindp3 = 0.01;
    double Maxdp3 = 1.0;

	IntOpt("seed", &Seed);
	IntOpt("paths", &PathCount);

	IntOpt("subset1", &SubsetSize1);
	IntOpt("subset2", &SubsetSize2);

	IntOpt("maxdirs2", &MaxDirs2);
	FloatOpt("mindq2", &Mindq2);
	FloatOpt("maxdq2", &Maxdq2);
	FloatOpt("mindp2", &Mindp2);
	FloatOpt("maxdp2", &Maxdp2);
	bool Fast2 = true;

	IntOpt("maxdirs3", &MaxDirs3);
	FloatOpt("mindq3", &Mindq3);
	FloatOpt("maxdq3", &Maxdq3);
	FloatOpt("mindp3", &Mindp3);
	FloatOpt("maxdp3", &Maxdp3);
	bool Fast3 = FlagOpt("fast3");

	InitRand(Seed);

	PairWiseAligner::InitFromCmdLine();
	GetSubstMxFromCmdLine();

	PWSet Set1;
	MSASet M;
	M.FromFile(InFileName);
	M.GetPWSubset(SubsetSize1, Set1);
	g_ptrPWSet = &Set1;

	Log("\n");
	Log("PWA=%s\n", g_PWA->GetName());
	ParamVec::FixMF(true);

	Gridder G;
	G.InitFromCmdLine();
	
	SetLogProgress(false);

	G.ComputeGrid(PWGetQ);

	vector<uns> CellIndexesByQ;
	InitSortOrder(CellIndexesByQ, G.GetCellCount());
	G.SortCellIndexesByQ(CellIndexesByQ);

	if (CellIndexesByQ.size() > 8)
		CellIndexesByQ.resize(8);

	reverse(CellIndexesByQ.begin(), CellIndexesByQ.end());
	Log("\n");
	Log("Best:\n");
	G.LogCells(CellIndexesByQ);
	reverse(CellIndexesByQ.begin(), CellIndexesByQ.end());

	vector<uns> StartCellIndexes;
	G.GetLocalOptima(StartCellIndexes);
	G.SortCellIndexesByQ(StartCellIndexes);
	const uns N = SIZE(StartCellIndexes);
	Log("\n");
	Log("%u local optima:\n", N);
	G.LogCells(StartCellIndexes);

	if (SIZE(StartCellIndexes) < PathCount)
		FillStartCells(StartCellIndexes, PathCount, CellIndexesByQ);

	PWSet Set2;
	M.GetPWSubset(SubsetSize1, Set2);
	g_ptrPWSet = &Set2;

	PathCount = min(PathCount, SIZE(StartCellIndexes));

	vector<ParamVec> Starts3;
	for (uns i = 0; i < PathCount; ++i)
		{
		uns CellIndex = StartCellIndexes[i];
		const ParamVec &StartParams = G.m_ParamVecs[CellIndex];

		ParamVec StartGradients;
		G.GetGradient(CellIndex, StartGradients);

		HillClimber H;
		H.m_MindQ = Mindq2;
		H.m_MaxdQ = Maxdq2;
		H.m_MindParam = Mindp2;
		H.m_MaxdParam = Maxdp2;
		H.m_Fast = Fast2;

		Log("\n");
		Log("Iter 2 hillclimb Params=");
		StartParams.LogMe();
		Log(" Grads=");
		StartGradients.LogMe();
		Log("\n");
		Log("\n");
		H.Climb(PWGetQ, StartParams, StartGradients, MaxDirs2, "POP2");

		Log("\n");
		Log("Iter 2 Path %u BestQ=%7.5f Params=", i, H.m_BestQ);
		H.m_BestParams.LogMe();
		Log("\n");

		Starts3.push_back(H.m_BestParams);
		}

	PWSet Set3;
	M.GetAllPairs(Set3);
	M.Clear();
	g_ptrPWSet = &Set3;

	vector<ParamVec> FinalParamVec;
	vector<double> FinalQs;

	double BestQ = -1;
	ParamVec BestParams;

	for (uns i = 0; i < PathCount; ++i)
		{
		const ParamVec &StartParams = Starts3[i];

		ParamVec StartGradients;
		uns CellIndex = StartCellIndexes[i];
		G.GetGradient(CellIndex, StartGradients);

		HillClimber H;
		H.m_MindQ = Mindq3;
		H.m_MaxdQ = Maxdq3;
		H.m_MindParam = Mindp3;
		H.m_MaxdParam = Maxdp3;
		H.m_Fast = Fast3;

		Log("\n");
		Log("Iter 3 hillclimb Params=");
		StartParams.LogMe();
		Log(" Grads=");
		StartGradients.LogMe();
		Log("\n");
		Log("\n");
		H.Climb(PWGetQ, StartParams, StartGradients, MaxDirs3, "POP3");

		FinalParamVec.push_back(H.m_BestParams);
		FinalQs.push_back(H.m_BestQ);

		if (H.m_BestQ > BestQ)
			{
			BestQ = H.m_BestQ;
			BestParams = H.m_BestParams;
			}

		Log("\n");
		Log("Iter 3 Path %u BestQ=%7.5f Params=", i, H.m_BestQ);
		H.m_BestParams.LogMe();
		Log("\n");
		}
	Log("\n");
	Log("\n");

	vector<uns> SortOrder;
	InitSortOrder(SortOrder, PathCount);
	GetSortOrder(FinalQs, SortOrder);

	Log("Final optima:\n");
	ParamVec::LogRowHdr1();
	Log("        Q");
	Log("\n");
	ParamVec::LogRowHdr2();
	Log("  -------");
	Log("\n");
	reverse(SortOrder.begin(), SortOrder.end());
	for (uns i = 0; i < PathCount; ++i)
		{
		uns j = SortOrder[i];
		const ParamVec &Params = FinalParamVec[j];
		Params.LogRow();
		double Q = FinalQs[j];
		Log("  %7.5f\n", Q);
		}
	Log("\n");
	string tmp;
	string Opts;
	Log("Algo=pop;Set=%s;Mx=%s;Q=%7.5f;%sCmdLine=%s\n",
	  InFileName,
	  RequiredValueOpt("mx"),
	  BestQ,
	  BestParams.ToSepStr(Opts).c_str(),
	  GetCmdLine(tmp).c_str());
	}
