#include "mus.h"

uns UngappedSeqLength(const string &s);

static void TruncSeq(const string &s, string &t, bool &LeftTrunc, bool &RightTrunc)
	{
	t.clear();

	LeftTrunc = false;
	RightTrunc = false;

	bool AnyLetters = false;
	for (string::const_iterator p = s.begin(); p != s.end(); ++p)
		{
		char c = *p;
		if (isalpha(c))
			AnyLetters = true;
		else if (c == '+')
			{
			if (AnyLetters)
				RightTrunc = true;
			else
				LeftTrunc = true;
			c = '-';
			}
		t.push_back(c);
		}
	}

PWSet::PWSet()
	{
	Clear();
	}

void PWSet::Clear()
	{
	m_Labels.clear();
	m_PairNames.clear();
	m_PairWeights.clear();
	m_LabelPairs.clear();
	m_Names.clear();
	m_GappedSeqs.clear();
	m_UngappedSeqs.clear();
	m_GappedBVs.clear();
	m_UngappedBVs.clear();
	m_RefPWAs.clear();
	m_UngappedBAs.clear();
	m_HillClimbBestParams.Clear();
	m_HillClimbBestQ = 0;
	m_HillQs.clear();
	m_HillParams.clear();
	m_MSAIndexes.clear();
	m_MSASeqIndexes.clear();
	m_TrainMaxIters = 0;
	m_TrainInitIters = 0;
	m_TrainLearnIters = 0;
	m_TrainConvergeIters = 0;
	m_LogIterFrom = UINT_MAX;
	m_LogIterTo = UINT_MAX;
	m_TrainDeltaWeightStyle = dWPerAln;
	m_MinQImprovement = 0;
	m_TrainTrace = false;
	m_TrainRandomizeOrder = false;
	m_Iter = 0;
	m_TrainBestIter = 0;
	m_TrainCurrParams.Clear();
	m_TrainBestParams.Clear();
	m_TrainSumGradientsCurrIter.Clear();
	m_LossScale = 0;
	m_TrainCurrQ = 0;
	m_TrainBestQ = 0;
	m_SumPairWeights = 0;
	m_MemSave = false;
	}

void SeqsToByteVecs(const vector<string> &GappedSeqs, vector<string> &UngappedSeqs,
  vector<vector<byte> > &GappedBVs, vector<vector<byte> > &UngappedBVs,
  vector<byte *> &UngappedByteArrays)
	{
	UngappedSeqs.clear();
	GappedBVs.clear();
	UngappedBVs.clear();
	const uns C = SIZE(UngappedByteArrays);
	for (uns i = 0; i < C; ++i)
		delete[] UngappedByteArrays[i];
	UngappedByteArrays.clear();

	size_t N = GappedSeqs.size();
	UngappedSeqs.resize(N);
	GappedBVs.resize(N);
	UngappedBVs.resize(N);
	UngappedByteArrays.resize(N);

	ProgressStart("%u pairs, converting to vectors", N/2);
	for (size_t i = 0; i < N; ++i)
		{
		if (i%100 == 0)
			ProgressStep((int) i, (int) N);
		const string &GappedSeq = GappedSeqs[i];

		string &UngappedSeq = UngappedSeqs[i];

		vector<byte> &GappedBV = GappedBVs[i];
		vector<byte> &UngappedBV = UngappedBVs[i];

		StripGaps(GappedSeq, UngappedSeq);
		StrToLetters(GappedSeq, GappedBV);
		StripGaps(GappedBV, UngappedBV);

		const uns L = SIZE(UngappedSeq);
		byte *UngappedBA = new byte[L];
		for (uns j = 0; j < L; ++j)
			UngappedBA[j] = UngappedBV[j];
		UngappedByteArrays[i] = UngappedBA;
		}
	ProgressDone();
	}

void PWSet::GetRefPair(uns PairIndex, RefPairData &Ref) const
	{
	if (PairIndex >= SIZE(m_RefPWAs))
		Die("GetRefPair(%u), there are %u alns", PairIndex, SIZE(m_RefPWAs));

	uns iA = 2*PairIndex;
	uns iB = iA + 1;

	Ref.LabelA = &m_Labels[iA];
	Ref.LabelB = &m_Labels[iB];

	Ref.GSA = &m_GappedSeqs[iA];
	Ref.GSB = &m_GappedSeqs[iB];
	Ref.USA = &m_UngappedSeqs[iA];
	Ref.USB = &m_UngappedSeqs[iB];

	Ref.GVA = &m_GappedBVs[iA];
	Ref.GVB = &m_GappedBVs[iB];
	Ref.UVA = &m_UngappedBVs[iA];
	Ref.UVB = &m_UngappedBVs[iB];

	Ref.UGAA = m_UngappedBAs[iA];
	Ref.UGAB = m_UngappedBAs[iB];

	Ref.Labels = &m_LabelPairs[PairIndex];
	Ref.RefPWA = &m_RefPWAs[PairIndex];

	Ref.LA = SIZE(*Ref.USA);
	Ref.LB = SIZE(*Ref.USB);

	Ref.Name = &m_PairNames[PairIndex];

	Ref.Weight = m_PairWeights[PairIndex];

	Ref.MSAIndex = m_MSAIndexes[PairIndex];
	Ref.SeqIndex1 = m_MSASeqIndexes[PairIndex].first;
	Ref.SeqIndex2 = m_MSASeqIndexes[PairIndex].second;

	Ref.ALeftTrunc = m_LeftTruncs[iA];
	Ref.ARightTrunc = m_RightTruncs[iA];
	Ref.BLeftTrunc = m_LeftTruncs[iB];
	Ref.BRightTrunc = m_RightTruncs[iB];
	}

static PWSet *g_ptrPWSet;

double GetWeightFromLabel(const string &Label)
	{
	size_t pos = Label.find(";Weight=");
	if (pos != string::npos)
		{
		string w = Label.substr(pos+8);
		return atof(w.c_str());
		}

	const char *s = Label.c_str();
	uns L = SIZE(Label);
	if (L < 3)
		return 1.0;
	if (Label[L-1] != ')')
		return 1.0;
	string strN;
	for (int i = (int) L - 2; i >= 0; --i)
		{
		char c = Label[i];
		if (c == '(')
			{
			if (i == 0)
				Die("GetWeightFromLabel(%s), invalid", s);
			if (Label[i-1] != 'N')
				return 1.0;
			break;
			}
		if (i == 0)
			Die("GetWeightFromLabel(%s), invalid", s);
		strN = c + strN;
		}
	int N = atoi(strN.c_str());
	if (N == 0)
		Die("GetWeightFromLabel(%s), N=0 invalid", s);
	int PairCount = (N*(N-1))/2;
	return 1.0/PairCount;
	}

static void OnSeq(const string &Label, const string &Seq)
	{
	g_ptrPWSet->m_Labels.push_back(Label);
	g_ptrPWSet->m_GappedSeqs.push_back(Seq);
	}

static void OnGroupLabel(const string &Label)
	{
	g_ptrPWSet->m_PairNames.push_back(Label);
	double Weight = GetWeightFromLabel(Label);
	g_ptrPWSet->m_PairWeights.push_back(Weight);
	}

void PWSet::ToFile(const char *FileName) const
	{
	FILE *f = CreateStdioFile(FileName);
	const uns N = SIZE(m_GappedSeqs);
	asserta(SIZE(m_Labels) == N);
	for (uns i = 0; i < N; ++i)
		{
		const string &Label = m_Labels[i];
		const string &Seq = m_GappedSeqs[i];
		fprintf(f, ">%s\n", Label.c_str());
		WriteFasta(f, Seq.c_str(), SIZE(Seq));
		}
	CloseStdioFile(f);
	}

void PWSet::FromFile(const char *FileName, bool MemSave)
	{
	Clear();
	m_MemSave = MemSave;
	g_ptrPWSet = this;
	ReadFasta(FileName, OnSeq, OnGroupLabel);
	Progress("%.32s, %u seqs", FileName, SIZE(m_GappedSeqs));
	uns N = SIZE(m_GappedSeqs)/2;
	uns M = SIZE(m_PairNames);
	if (M == 0)
		{
		if (N == 1)
			{
			m_PairNames.push_back("");
			m_PairWeights.push_back(1.0);
			M = 1;
			}
		else
			Die("%.32s, pair names not specified", FileName);
		}
	if (M != N)
		Die("%.32s, %u pair names but %u sequence pairs", FileName, M, N);
	Init();
	}

void PWSet::Init()
	{
	const uns N = SIZE(m_GappedSeqs);
	if (N == 0 || N%2 != 0)
		Die("Need x2 seqs in training set, got %u", (uns) N);

	double SumPairWeights = ComputeSumPairWeights();
	double x = (SumPairWeights == 0 ? 1 : N/SumPairWeights);

	vector<string> EmptyStrPair(2);
	const uns PairCount = N/2;
	for (uns PairIndex = 0; PairIndex < PairCount; ++PairIndex)
		{
		uns i = 2*PairIndex;
		uns i1 = i + 1;
		const string &LabelA = m_Labels[i];
		const string &LabelB = m_Labels[i1];

		const string &GA = m_GappedSeqs[i];
		const string &GB = m_GappedSeqs[i1];
		uns LA = SIZE(GA);
		uns LB = SIZE(GB);
		if (LA != LB)
			Die("Seqs %.32s,%.32s not aligned (lengths %u, %u)",
			  m_Labels[i].c_str(),
			  m_Labels[i1].c_str(),
			  LA, LB);

		m_RefPWAs.push_back(EmptyStrPair);
		m_LabelPairs.push_back(EmptyStrPair);
		m_MSAIndexes.push_back(0);
		m_MSASeqIndexes.push_back(pair<unsigned, unsigned>(0, 1));

		vector<string> &RefPWA = m_RefPWAs.back();

		string GAt;
		string GBt;
		bool ALeftTrunc;
		bool ARightTrunc;
		bool BLeftTrunc;
		bool BRightTrunc;
		TruncSeq(GA, GAt, ALeftTrunc, ARightTrunc);
		TruncSeq(GB, GBt, BLeftTrunc, BRightTrunc);
		RefPWA[0] = GAt;
		RefPWA[1] = GBt;
		m_LeftTruncs.push_back(ALeftTrunc);
		m_RightTruncs.push_back(ARightTrunc);
		m_LeftTruncs.push_back(BLeftTrunc);
		m_RightTruncs.push_back(BRightTrunc);

		vector<string> &LabelPair = m_LabelPairs.back();
		LabelPair[0] = LabelA;
		LabelPair[1] = LabelB;

		m_Names.push_back(m_PairNames[PairIndex]);

		if (SumPairWeights == 0)
			m_PairWeights[PairIndex] = 1.0;
		else
			m_PairWeights[PairIndex] *= x;
		}

	if (m_MemSave)
		return;

	SeqsToByteVecs(m_GappedSeqs, m_UngappedSeqs, m_GappedBVs, m_UngappedBVs,
	  m_UngappedBAs);
	}

void PWSet::AddPair(const string &Name, const string &LabelA, const string &SeqA,
  const string &LabelB, const string &SeqB, double Weight, unsigned MSAIndex,
  unsigned SeqIndex1, unsigned SeqIndex2)
	{
	uns ULA = UngappedSeqLength(SeqA);
	uns ULB = UngappedSeqLength(SeqB);
	if (ULA*ULB > 1000000)
		{
		Log("Pair %s,%s LA=%u, LB=%u discarded (too big)\n",
		  LabelA.c_str(), LabelB.c_str(), ULA, ULB);
		return;
		}
	m_PairNames.push_back(Name);
	m_Labels.push_back(LabelA);
	m_Labels.push_back(LabelB);
	m_PairWeights.push_back(Weight);
	m_MSAIndexes.push_back(MSAIndex);
	m_MSASeqIndexes.push_back(pair<uns, uns>(SeqIndex1, SeqIndex2));

	string s;
	bool LeftTrunc;
	bool RightTrunc;
	TruncSeq(SeqA, s, LeftTrunc, RightTrunc);
	m_GappedSeqs.push_back(s);
	m_LeftTruncs.push_back(LeftTrunc);
	m_RightTruncs.push_back(RightTrunc);

	TruncSeq(SeqB, s, LeftTrunc, RightTrunc);
	m_GappedSeqs.push_back(s);
	m_LeftTruncs.push_back(LeftTrunc);
	m_RightTruncs.push_back(RightTrunc);
	}

double PWSet::ComputeSumPairWeights() const
	{
	const uns N = SIZE(m_PairWeights);
	double Sum = 0;
	for (uns i = 0; i < N; ++i)
		Sum += m_PairWeights[i];
	return Sum;
	}

void PWSet::GetMSACoords(uns PairIndex, uns &MSAIndex, uns &SeqIndex1,
  uns &SeqIndex2) const
	{
	asserta(PairIndex < SIZE(m_MSAIndexes));
	asserta(SIZE(m_MSAIndexes) == SIZE(m_MSASeqIndexes));
	MSAIndex = m_MSAIndexes[PairIndex];
	SeqIndex1 = m_MSASeqIndexes[PairIndex].first;
	SeqIndex2 = m_MSASeqIndexes[PairIndex].second;
	}
