#ifndef pwset_h
#define pwset_h

#include "paramspace.h"

class MSASet;

struct RefPairData
	{
	uns LA;
	uns LB;
	const string *Name;
	const string *LabelA;
	const string *LabelB;
	const vector<string> *Labels;
	const string *GSA;
	const string *GSB;
	const string *USA;
	const string *USB;
	const byte *UGAA;
	const byte *UGAB;
	const vector<byte> *GVA;
	const vector<byte> *GVB;
	const vector<byte> *UVA;
	const vector<byte> *UVB;
	const vector<string> *RefPWA;
	double Weight;
	uns MSAIndex;
	uns SeqIndex1;
	uns SeqIndex2;
	bool ALeftTrunc;
	bool ARightTrunc;
	bool BLeftTrunc;
	bool BRightTrunc;

	const char *NameA()
		{
		return LabelA->c_str();
		}

	const char *NameB()
		{
		return LabelB->c_str();
		}

	const string &GetName()
		{
		return *Name;
		}

	void LogMe() const
		{
		Log("\n");
		Log("RefPairData Name=%s LA=%u LB=%u Weight=%g\n",
		  Name->c_str(), LA, LB, Weight);
		}
	};

enum DeltaWeightStyle
	{
	dWPerAln,
	dWPerIter,
	dWAlternate,
	};

typedef SCORE (*ptrPWAlnFunc)(const vector<byte> &A, const vector<byte> &B,
  string &Path);

class PWSet
	{
public:
	PWSet();
	void Clear();
	void SetMemSave(bool On = true) { m_MemSave = On; }
	void FromFile(const char *FileName, bool MemSave = false);
	void FromFileM(const char *FileName, MSASet &M, uns SubsetSize = UINT_MAX);
	void ToFile(const char *FileName) const;
	void Init();
	unsigned GetPairCount() const
		{
		return SIZE(m_PairNames);
		};
	double GetWeight(uns PairIndex) const;

	void AddPair(const string &Name, const string &LabelA, const string &SeqA,
	  const string &LabelB, const string &SeqB, double Weight = 1.0,
	  unsigned MSAIndex = UINT_MAX, unsigned SeqIndex1 = UINT_MAX,
	  unsigned SeqIndex2 = UINT_MAX);
	const string &GetPairName(uns PairIndex) const
		{
		asserta(PairIndex < SIZE(m_Names));
		return m_Names[PairIndex];
		}
	double GetSumPairWeights() const
		{
		return m_SumPairWeights;
		}
	double ComputeSumPairWeights() const;
	void GetMSACoords(uns PairIndex, uns &MSAIndex, uns &SeqIndex1,
	  uns &SeqIndex2) const;

	void GetRefPair(uns PairIndex, RefPairData &Ref) const;
	void TrainParamsFromCmdLine();
	double Train(ParamVec &BestParams);
	void TrainPerSample();
	void TrainPerSet();
	void LogTrainIter() const;
	void LogTrainIterHdr() const;
	void LogTrainParams() const;
	void TrainPair(uns PairIndex, DeltaWeightStyle dWStyle);

	double ComputeQ(const ParamVec &Params);
	double ComputeQPair(uns PairIndex, const ParamVec &Params) const;
	double ComputeQs(const ParamVec &Params, vector<double> &Qs) const;

	void HillClimbLogGetQ(const ParamVec &Params, double Q,
	  const string &Why = "") const;
	void HillCimbLogGetQHdr() const;
	double HillClimbGetQ(const ParamVec &Params, const string &Why = "");
	void HillClimb(const ParamVec &InitParams, const ParamVec &InitGradients);
	bool HillClimbIter(const ParamVec &StartParams, ParamVec &Gradients,
	  ParamVec &BestParams, double &BestQ);
	SCORE GetHillClimbDeltaParam(double CenterQ, const ParamVec &InitParams,
	  uns ParamIndex, SCORE Gradient, bool Plus);
	SCORE GetImprovedDeltaParam(double CenterQ, double InitQ, 
	  const ParamVec &InitParams, uns ParamIndex, SCORE InitDelta);
	void GetHillClimbParamDeltas(const ParamVec &Params,
	  const ParamVec &Gradients, double Q, ParamVec &PlusDeltas,
	  ParamVec &MinusDeltas);
	//void FindAnchors(ptrPWAlnFunc AlnFunc);
	//void GetAnchorsPair(uns PairIndex, ptrPWAlnFunc AlnFunc, 
	//  vector<pair<uns, uns> > &Anchors) const;

public:
	vector<string> m_Labels;
	vector<string> m_PairNames;
	vector<double> m_PairWeights;
	vector<vector<string> > m_LabelPairs;
	vector<string> m_Names;
	vector<string> m_GappedSeqs;
	vector<string> m_UngappedSeqs;
	vector<vector<byte> > m_GappedBVs;
	vector<vector<byte> > m_UngappedBVs;
	vector<vector<string> > m_RefPWAs;
	vector<byte *> m_UngappedBAs;
	vector<bool> m_LeftTruncs;
	vector<bool> m_RightTruncs;
//	vector<vector<pair<uns, uns> > > m_Anchors;

// Hill climbing state
	ParamVec m_HillClimbBestParams;
	double m_HillClimbBestQ;
	vector<double> m_HillQs;
	vector<ParamVec> m_HillParams;

// Coordinates in MSA set, if applicable
	vector<uns> m_MSAIndexes;
	vector<pair<uns, uns> > m_MSASeqIndexes;

// Train parameters
	uns m_TrainMaxIters;
	uns m_TrainInitIters;
	uns m_TrainLearnIters;
	uns m_TrainConvergeIters;
	uns m_LogIterFrom;
	uns m_LogIterTo;
	DeltaWeightStyle m_TrainDeltaWeightStyle;
	double m_MinQImprovement;
	bool m_TrainTrace;
	bool m_TrainRandomizeOrder;
	bool m_MemSave;

// Training state
	uns m_Iter;
	uns m_TrainBestIter;
	ParamVec m_TrainCurrParams;
	ParamVec m_TrainBestParams;
	ParamVec m_TrainSumGradientsCurrIter;
	double m_LossScale;
	double m_TrainCurrQ;
	double m_TrainBestQ;
	double m_SumPairWeights;
	double m_TrainSumQ;
	bool m_Improved;
	};

#endif // pwset_h
