#define TRACE	0

#include "mus.h"

// O(NL) computation of PREFAB Q score and Balibase TC score.
// Algorithm based on an idea due to Chuong (Tom) Do.
// Each position in the reference alignment is annotated with
// the column number C in the test alignment where the same 
// letter is found. A pair of identical Cs in the same reference
// column indicates a correctly aligned pair of letters.
void QScore(const char *RefName, 
  const vector<string> &msaTest, const vector<string> &TestLabels,
  const vector<string> &msaRef,  const vector<string> &RefLabels,
  double &Q, double &TC, bool WarnIfNoRefAligned)
	{
	Q = 0;
	TC = 0;

	unsigned CorrectPairCount = 0;
	unsigned RefAlignedPairCount = 0;

	const unsigned RefSeqCount = SIZE(msaRef);
	const unsigned TestSeqCount = SIZE(msaTest);

	if (RefSeqCount == 0 || TestSeqCount == 0)
		{
		Warning("QScore(%.32s), test has %u seqs, ref has %u seqs",
		  RefName, TestSeqCount, RefSeqCount);
		return;
		}

	const unsigned RefColCount = SIZE(msaRef[0]);
	const unsigned TestColCount = SIZE(msaTest[0]);

	map<string, unsigned> RefSeqNameToIndex;
	vector<unsigned> RefToTestSeqIndex(RefSeqCount);

	for (unsigned RefSeqIndex = 0; RefSeqIndex < RefSeqCount; ++RefSeqIndex)
		{
		const string &SeqName = RefLabels[RefSeqIndex];
		RefToTestSeqIndex[RefSeqIndex] = UINT_MAX;
		RefSeqNameToIndex[SeqName] = RefSeqIndex;
		}

	for (unsigned TestSeqIndex = 0; TestSeqIndex < TestSeqCount; ++TestSeqIndex)
		{
		const string &SeqName = TestLabels[TestSeqIndex];
		map<string, unsigned>::const_iterator p = RefSeqNameToIndex.find(SeqName);
		if (p != RefSeqNameToIndex.end())
			{
			unsigned RefSeqIndex = p->second;
			if (RefSeqIndex == UINT_MAX)
				Die("QScore, RefSeq %.32s not in map", SeqName.c_str());
			RefToTestSeqIndex[RefSeqIndex] = TestSeqIndex;
			}
		}

// TestColIndex[i] is the one-based (not zero-based!) test column index
// of the letter found in the current column of the reference alignment
// (or the most recent letter if the reference column is gapped, or zero
// if no letter has yet been found). Here, seq index i is for msaRef.
	vector<unsigned> TestColIndex(TestSeqCount, 0);

// TestColIndexCount[i] is the number of times that a letter from test
// column i (one-based!) appears in the current reference column.
	vector<unsigned> TestColIndexCount(TestColCount+1, 0);

// TestColIndexes[i] is the column index in the test alignment of
// the i'th non-gapped position in the current reference column.
	vector<unsigned> TestColIndexes;

	unsigned RefAlignedColCount = 0;
	unsigned CorrectColCount = 0;
	for (unsigned RefColIndex = 0; RefColIndex < RefColCount; RefColIndex++)
		{
#if	TRACE
		Log("\n");
		Log("Ref col %u\n", RefColIndex);
#endif
		TestColIndexes.clear();
		TestColIndexes.reserve(RefSeqCount);

	// NonGappedCount is the number of non-gapped positions in the current
	// reference column.
		unsigned NonGappedCount = 0;
		unsigned FirstTestColIndex = UINT_MAX;
		bool RefColIsAligned = false;
		bool TestColAllCorrect = true;
		for (unsigned RefSeqIndex = 0; RefSeqIndex < RefSeqCount; RefSeqIndex++)
			{
			unsigned TestSeqIndex = RefToTestSeqIndex[RefSeqIndex];
			if (TestSeqIndex == UINT_MAX)
				Die("Test seq %.32s missing", RefLabels[RefSeqIndex].c_str());
#if	TRACE
			Log("  Seq %.32s, test seq ix %u\n", RefLabels[RefSeqIndex].c_str(), TestSeqIndex);
#endif

			char cRef = msaRef[RefSeqIndex][RefColIndex];
#if	TRACE
			Log("  cRef=%c\n", cRef);
#endif
			if (IsGapChar(cRef))
				continue;

			char cTest = 0;
			unsigned Col = TestColIndex[TestSeqIndex];
			do
				{
				if (Col >= TestColCount)
					Die("QScore, past end-of-seq in test");
				cTest = msaTest[TestSeqIndex][Col];
#if	TRACE
				Log("  Test[%u][%u] = %c\n", TestSeqIndex, Col, cTest);
#endif
				++Col;
				}
			while (IsGapChar(cTest));
			if (toupper(cRef) != toupper(cTest))
				Die("Seq %.32s different in ref & test alns, ref col %u=%c, test=%c",
				  RefLabels[RefSeqIndex].c_str(), RefColIndex, cRef, cTest);

			if (isupper(cRef))
				{
				RefColIsAligned = true;
				++NonGappedCount;
				if (isupper(cTest))
					{
					TestColIndexes.push_back(Col);
					++(TestColIndexCount[Col]);
					}
				if (FirstTestColIndex == UINT_MAX)
					FirstTestColIndex = Col;
				else
					{
					if (FirstTestColIndex != Col)
						TestColAllCorrect = false;
					}
				}
			else
				{
				if (RefColIsAligned)
					Die("Ref %s, col %u has both upper- and lower-case letters",
					  RefName, RefColIndex);
				}

			TestColIndex[TestSeqIndex] = Col;
			}

		if (RefColIsAligned && NonGappedCount > 1)
			{
			++RefAlignedColCount;
			if (TestColAllCorrect)
				++CorrectColCount;
			}

		unsigned ColPairCount = 0;
		for (vector<unsigned>::const_iterator p = TestColIndexes.begin();
		  p != TestColIndexes.end(); ++p)
			{
			unsigned Col = *p;
			unsigned Count = TestColIndexCount[Col];
			if (Count > 0)
				ColPairCount += Count*(Count - 1)/2;
			TestColIndexCount[Col] = 0;
			}

		CorrectPairCount += ColPairCount;
		RefAlignedPairCount += NonGappedCount*(NonGappedCount - 1)/2;
		}

	if (RefAlignedPairCount == 0)
		Q = 0;
	else
		Q = (double) CorrectPairCount / (double) RefAlignedPairCount;

	if (RefAlignedColCount == 0)
		{
		if (WarnIfNoRefAligned)
			Warning("reference %s has no aligned (upper-case) columns", RefName);
		TC = 0;
		}
	else
		TC = (double) CorrectColCount / (double) RefAlignedColCount;
	}
